@@ -280,10 +280,12 @@ void init_layer3(void)
 		int *mp;
 		int cb,lwin;
 		const unsigned char *bdf;
+		int switch_idx;
 
 		mp = map[j][0] = mapbuf0[j];
 		bdf = bi->longDiff;
-		for(i=0,cb = 0; cb < 8 ; cb++,i+=*bdf++)
+		switch_idx = (j < 3) ? 8 : 6;
+		for(i=0,cb = 0; cb < switch_idx ; cb++,i+=*bdf++)
 		{
 			*mp++ = (*bdf) >> 1;
 			*mp++ = i;
@@ -465,8 +467,14 @@ static int III_get_side_info(mpg123_handle *fr, struct III_sideinfo *si,int ster
 		}
 		gr_info->pow2gain = fr->gainpow2+256 - getbits_fast(fr, 8) + powdiff;
 		if(ms_stereo) gr_info->pow2gain += 2;
-
 		gr_info->scalefac_compress = getbits(fr, tab[4]);
+		if(gr_info->part2_3_length == 0)
+		{
+			if(gr_info->scalefac_compress > 0)
+				debug1( "scalefac_compress _should_ be zero instead of %i"
+				,	gr_info->scalefac_compress );
+			gr_info->scalefac_compress = 0;
+		}
 
 		if(get1bit(fr))
 		{ /* window switch flag  */
@@ -553,6 +561,14 @@ static int III_get_scale_factors_1(mpg123_handle *fr, int *scf,struct gr_info_s
 	int num0 = slen[0][gr_info->scalefac_compress];
 	int num1 = slen[1][gr_info->scalefac_compress];
 
+	if(gr_info->part2_3_length == 0)
+	{
+		int i;
+		for(i=0;i<39;i++)
+			*scf++ = 0;
+		return 0;
+	}
+
 	if(gr_info->block_type == 2)
 	{
 		int i=18;
@@ -667,6 +683,14 @@ static int III_get_scale_factors_2(mpg123_handle *fr, int *scf,struct gr_info_s
 
 	pnt = stab[n][(slen>>12)&0x7];
 
+	if(gr_info->part2_3_length == 0)
+	{
+		int i;
+		for(i=0;i<39;i++)
+			*scf++ = 0;
+		return 0;
+	}
+
 	for(i=0;i<4;i++)
 	{
 		int num = slen & 0x7;
@@ -680,7 +704,7 @@ static int III_get_scale_factors_2(mpg123_handle *fr, int *scf,struct gr_info_s
 		else
 		for(j=0;j<(int)(pnt[i]);j++) *scf++ = 0;
 	}
-  
+
 	n = (n << 1) + 1;
 	for(i=0;i<n;i++) *scf++ = 0;
 
@@ -735,6 +759,12 @@ static int III_dequantize_sample(mpg123_handle *fr, real xr[SBLIMIT][SSLIMIT],in
 	int gainpow2_scale_idx = 378;
 #endif
 
+	/* Assumption: If there is some part2_3_length at all, there should be
+	   enough of it to work with properly. In case of zero length we silently
+	   zero things. */
+	if(gr_info->part2_3_length > 0)
+	{
+
 	/* mhipp tree has this split up a bit... */
 	int num=getbitoffset(fr);
 	MASK_TYPE mask;
@@ -771,7 +801,14 @@ static int III_dequantize_sample(mpg123_handle *fr, real xr[SBLIMIT][SSLIMIT],in
 			}
 		}
 	}
- 
+
+#define CHECK_XRPNT if(xrpnt >= &xr[SBLIMIT][0]) \
+{ \
+	if(NOQUIET) \
+		error2("attempted xrpnt overflow (%p !< %p)", (void*) xrpnt, (void*) &xr[SBLIMIT][0]); \
+	return 1; \
+}
+
 	if(gr_info->block_type == 2)
 	{
 		/* decoding with short or mixed mode BandIndex table */
@@ -852,6 +889,7 @@ static int III_dequantize_sample(mpg123_handle *fr, real xr[SBLIMIT][SSLIMIT],in
 					y &= 0xf;
 #endif
 				}
+				CHECK_XRPNT;
 				if(x == 15 && h->linbits)
 				{
 					max[lwin] = cb;
@@ -876,6 +914,7 @@ static int III_dequantize_sample(mpg123_handle *fr, real xr[SBLIMIT][SSLIMIT],in
 				else *xrpnt = DOUBLE_TO_REAL(0.0);
 
 				xrpnt += step;
+				CHECK_XRPNT;
 				if(y == 15 && h->linbits)
 				{
 					max[lwin] = cb;
@@ -908,22 +947,7 @@ static int III_dequantize_sample(mpg123_handle *fr, real xr[SBLIMIT][SSLIMIT],in
 			const struct newhuff* h;
 			const short* val;
 			register short a;
-			/*
-				This is only a humble hack to prevent a special segfault.
-				More insight into the real workings is still needed.
-				Especially why there are (valid?) files that make xrpnt exceed the array with 4 bytes without segfaulting, more seems to be really bad, though.
-			*/
-			#ifdef DEBUG
-			if(!(xrpnt < &xr[SBLIMIT][0]))
-			{
-				if(VERBOSE) debug2("attempted soft xrpnt overflow (%p !< %p) ?", (void*) xrpnt, (void*) &xr[SBLIMIT][0]);
-			}
-			#endif
-			if(!(xrpnt < &xr[SBLIMIT][0]+5))
-			{
-				if(NOQUIET) error2("attempted xrpnt overflow (%p !< %p)", (void*) xrpnt, (void*) &xr[SBLIMIT][0]);
-				return 2;
-			}
+
 			h = htc+gr_info->count1table_select;
 			val = h->table;
 
@@ -970,6 +994,7 @@ static int III_dequantize_sample(mpg123_handle *fr, real xr[SBLIMIT][SSLIMIT],in
 					}
 					mc--;
 				}
+				CHECK_XRPNT;
 				if( (a & (0x8>>i)) )
 				{
 					max[lwin] = cb;
@@ -994,6 +1019,7 @@ static int III_dequantize_sample(mpg123_handle *fr, real xr[SBLIMIT][SSLIMIT],in
 			{
 				for(;mc > 0;mc--)
 				{
+					CHECK_XRPNT;
 					*xrpnt = DOUBLE_TO_REAL(0.0); xrpnt += 3; /* short band -> step=3 */
 					*xrpnt = DOUBLE_TO_REAL(0.0); xrpnt += 3;
 				}
@@ -1083,6 +1109,7 @@ static int III_dequantize_sample(mpg123_handle *fr, real xr[SBLIMIT][SSLIMIT],in
 #endif
 				}
 
+				CHECK_XRPNT;
 				if(x == 15 && h->linbits)
 				{
 					max = cb;
@@ -1106,6 +1133,7 @@ static int III_dequantize_sample(mpg123_handle *fr, real xr[SBLIMIT][SSLIMIT],in
 				}
 				else *xrpnt++ = DOUBLE_TO_REAL(0.0);
 
+				CHECK_XRPNT;
 				if(y == 15 && h->linbits)
 				{
 					max = cb;
@@ -1174,6 +1202,7 @@ static int III_dequantize_sample(mpg123_handle *fr, real xr[SBLIMIT][SSLIMIT],in
 					}
 					mc--;
 				}
+				CHECK_XRPNT;
 				if( (a & (0x8>>i)) )
 				{
 					max = cb;
@@ -1198,6 +1227,18 @@ static int III_dequantize_sample(mpg123_handle *fr, real xr[SBLIMIT][SSLIMIT],in
 	backbits(fr, num);
 	num = 0;
 
+	}
+	else
+	{
+		part2remain = 0;
+		/* Not entirely sure what good values are, must be > 0. */
+		gr_info->maxband[0] =
+		gr_info->maxband[1] =
+		gr_info->maxband[2] =
+		gr_info->maxbandl   = 0;
+		gr_info->maxb       = 1;
+	}
+
 	while(xrpnt < &xr[SBLIMIT][0]) 
 	*xrpnt++ = DOUBLE_TO_REAL(0.0);
 
